from ..base import DBTestCase
from ..example_project.tables import Band


class TestExists(DBTestCase):

    def test_exists(self):
        self.insert_rows()

        response = Band.exists().where(
            Band.name == 'Pythonistas'
        ).run_sync()

        self.assertTrue(response == True)



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from piccolo.table import Table


class Create(Query):
    """
    Creates a database table.
    """

    __slots__ = ("if_not_exists",)

    def __init__(self, table: t.Type[Table], if_not_exists: bool = False):
        super().__init__(table)
        self.if_not_exists = if_not_exists

    @property
    def querystrings(self) -> t.Sequence[QueryString]:
        prefix = "CREATE TABLE"
        if self.if_not_exists:
            prefix += " IF NOT EXISTS"

        base = f"{prefix} {self.table._meta.tablename}"
        columns = ", ".join(["{}" for i in self.table._meta.columns])
        query = f"{base} ({columns})"
        return [
            QueryString(
                query, *[i.querystring for i in self.table._meta.columns]
            )
        ]



from unittest import TestCase

from piccolo.commands.migration.base import BaseMigrationManager, Migration


class TestBaseMigrationManager(TestCase):
    def test_create_migration_table(self):
        self.assertTrue(BaseMigrationManager().create_migration_table())

    def tearDown(self):
        Migration.raw("DROP TABLE IF EXISTS migration").run_sync()



from __future__ import annotations
import asyncio
import os
import pprint
import typing as t

import click

from piccolo.migrations.tables import Migration
from piccolo.commands.migration.base import (
    BaseMigrationManager,
    MigrationModule,
    ConfigModule,
)


class ForwardsMigrationManager(BaseMigrationManager):
    def run_migrations(self, config_modules: t.List[ConfigModule]) -> None:
        already_ran = Migration.get_migrations_which_ran()
        print(f"Already ran:\n{already_ran}\n")

        for config_module in config_modules:
            migrations_folder = os.path.dirname(config_module.__file__)

            migration_modules: t.Dict[
                str, MigrationModule
            ] = self.get_migration_modules(migrations_folder)

            ids = self.get_migration_ids(migration_modules)
            print(f"Migration ids = {ids}")

            app_name = self.get_app_name(config_module)

            havent_run = sorted(set(ids) - set(already_ran))
            for _id in havent_run:
                migration_module = migration_modules[_id]
                asyncio.run(migration_module.forwards())

                print(f"Ran {_id}")
                Migration.insert().add(
                    Migration(name=_id, app_name=app_name)
                ).run_sync()

    def run(self):
        print("Running migrations ...")
        self.create_migration_table()

        config_modules = self.get_config_modules()

        print("Config Modules:")
        pprint.pprint(config_modules)
        print("\n")

        self.run_migrations(config_modules)


@click.command()
def forwards():
    """
    Runs any migrations which haven't been run yet, or up to a specific
    migration.
    """
    manager = ForwardsMigrationManager()
    manager.run()



from __future__ import annotations
from dataclasses import dataclass
import itertools
import typing as t

from piccolo.columns.base import Column, OnDelete, OnUpdate
from piccolo.columns.column_types import ForeignKey
from piccolo.query.base import Query
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from piccolo.table import Table


@dataclass
class AlterStatement:
    __slots__ = tuple()

    def querystring(self) -> QueryString:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclass
class AlterColumnStatement(AlterStatement):
    __slots__ = ("column",)

    column: t.Union[Column, str]

    @property
    def column_name(self) -> str:
        if isinstance(self.column, str):
            return self.column
        elif isinstance(self.column, Column):
            return self.column._meta.name
        else:
            raise ValueError("Unrecognised column type")


@dataclass
class RenameColumn(AlterColumnStatement):
    __slots__ = ("new_name",)

    new_name: str

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f"RENAME COLUMN {self.column_name} TO {self.new_name}"
        )


@dataclass
class DropColumn(AlterColumnStatement):
    @property
    def querystring(self) -> QueryString:
        return QueryString(f"DROP COLUMN {self.column_name}")


@dataclass
class AddColumn(AlterColumnStatement):
    __slots__ = ("name",)

    column: Column
    name: str

    @property
    def querystring(self) -> QueryString:
        self.column._meta.name = self.name
        return QueryString("ADD COLUMN {}", self.column.querystring)


@dataclass
class Unique(AlterColumnStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def querystring(self) -> QueryString:
        if self.boolean:
            return QueryString(f"ADD UNIQUE ({self.column_name})")
        else:
            if isinstance(self.column, str):
                raise ValueError(
                    "Removing a unique constraint requires a Column instance "
                    "to be passed as the column arg instead of a string."
                )
            tablename = self.column._meta.table._meta.tablename
            column_name = self.column_name
            key = f"{tablename}_{column_name}_key"
            return QueryString(f'DROP CONSTRAINT "{key}"')


@dataclass
class Null(AlterColumnStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def querystring(self) -> QueryString:
        if self.boolean:
            return QueryString(
                f"ALTER COLUMN {self.column_name} DROP NOT NULL"
            )
        else:
            return QueryString(f"ALTER COLUMN {self.column_name} SET NOT NULL")


@dataclass
class DropConstraint(AlterStatement):
    __slots__ = ("constraint_name",)

    constraint_name: str

    @property
    def querystring(self) -> QueryString:
        return QueryString(f"DROP CONSTRAINT IF EXISTS {self.constraint_name}")


@dataclass
class AddForeignKeyConstraint(AlterStatement):
    __slots__ = ("constraint_name",)

    constraint_name: str
    foreign_key_column_name: str
    referenced_table_name: str
    on_delete: t.Optional[OnDelete]
    on_update: t.Optional[OnUpdate]
    referenced_column_name: str = "id"

    @property
    def querystring(self) -> QueryString:
        query = (
            f"ADD CONSTRAINT {self.constraint_name} FOREIGN KEY "
            f"({self.foreign_key_column_name}) REFERENCES "
            f"{self.referenced_table_name} ({self.referenced_column_name})"
        )
        if self.on_delete:
            query += f" ON DELETE {self.on_delete.value}"
        if self.on_update:
            query += f" ON UPDATE {self.on_update.value}"
        return QueryString(query)


class Alter(Query):

    __slots__ = (
        "_add",
        "_drop",
        "_rename",
        "_unique",
        "_null",
        "_drop_table",
        "_drop_contraint",
        "_add_foreign_key_constraint",
    )

    def __init__(self, table: t.Type[Table]):
        super().__init__(table)
        self._add: t.List[AddColumn] = []
        self._drop: t.List[DropColumn] = []
        self._rename: t.List[RenameColumn] = []
        self._unique: t.List[Unique] = []
        self._null: t.List[Null] = []
        self._drop_table = False
        self._drop_contraint: t.List[DropConstraint] = []
        self._add_foreign_key_constraint: t.List[AddForeignKeyConstraint] = []

    def add_column(self, name: str, column: Column) -> Alter:
        """
        Band.alter().add_column(‘members’, Integer())
        """
        column._meta._table = self.table
        self._add.append(AddColumn(column, name))
        return self

    def drop_column(self, column: t.Union[str, Column]) -> Alter:
        """
        Band.alter().drop_column(Band.popularity)
        """
        self._drop.append(DropColumn(column))
        return self

    def drop_table(self) -> Alter:
        """
        Band.alter().drop_table()
        """
        self._drop_table = True
        return self

    def rename_column(
        self, column: t.Union[str, Column], new_name: str
    ) -> Alter:
        """
        Band.alter().rename_column(Band.popularity, ‘rating’)
        Band.alter().rename_column('popularity', ‘rating’)
        """
        self._rename.append(RenameColumn(column, new_name))
        return self

    def set_null(
        self, column: t.Union[str, Column], boolean: bool = True
    ) -> Alter:
        """
        Band.alter().set_null(Band.name, True)
        Band.alter().set_null('name', True)
        """
        self._null.append(Null(column, boolean))
        return self

    def set_unique(
        self, column: t.Union[str, Column], boolean: bool = True
    ) -> Alter:
        """
        Band.alter().set_unique(Band.name, True)
        Band.alter().set_unique('name', True)
        """
        self._unique.append(Unique(column, boolean))
        return self

    def _get_constraint_name(self, column: t.Union[str, ForeignKey]) -> str:
        column_name = AlterColumnStatement(column=column).column_name
        tablename = self.table._meta.tablename
        constraint_name = f"{tablename}_{column_name}_fk"
        return constraint_name

    def drop_constraint(self, constraint_name: str) -> Alter:
        self._drop_contraint.append(
            DropConstraint(constraint_name=constraint_name)
        )
        return self

    def drop_foreign_key_constraint(
        self, column: t.Union[str, ForeignKey]
    ) -> Alter:
        constraint_name = self._get_constraint_name(column=column)
        return self.drop_constraint(constraint_name=constraint_name)

    def add_foreign_key_constraint(
        self,
        column: t.Union[str, ForeignKey],
        referenced_table_name: str,
        on_delete: t.Optional[OnDelete] = None,
        on_update: t.Optional[OnUpdate] = None,
        referenced_column_name: str = "id",
    ) -> Alter:
        """
        This will add a new foreign key constraint.

        Band.alter().add_foreign_key_constraint(
            Band.manager,
            referenced_table_name='manager',
            on_delete=OnDelete.cascade
        )
        """
        constraint_name = self._get_constraint_name(column=column)
        column_name = AlterColumnStatement(column=column).column_name

        self._add_foreign_key_constraint.append(
            AddForeignKeyConstraint(
                constraint_name=constraint_name,
                foreign_key_column_name=column_name,
                referenced_table_name=referenced_table_name,
                on_delete=on_delete,
                on_update=on_update,
                referenced_column_name=referenced_column_name,
            )
        )
        return self

    async def response_handler(self, response):
        """
        We don't want to modify the response, but we need to add default values
        to any tables which have had columns added.
        """
        if self._add:
            # If the defaults are just static values then update all of the
            # rows in one go - otherwise we need to update each row at a time.
            # For example, with a UUID field, we need each row to get its
            # own unique value.

            columns = [i.column for i in self._add]

            defaults = [getattr(column, "default", None) for column in columns]

            just_static_defaults = all(
                [hasattr(default, "__call__") for default in defaults]
            )

            tablename = self.table._meta.tablename

            if just_static_defaults:
                for column in columns:
                    await self.table.raw(
                        f"UPDATE {tablename} SET {column._meta.name} = {{}}",
                        column.get_default_value(),
                    ).run()
            else:
                response = await self.table.raw(
                    f"SELECT id from {tablename}"
                ).run()
                ids = [i["id"] for i in response]

                for _id in ids:
                    for column in columns:
                        await self.table.raw(
                            (
                                f"UPDATE {tablename} "
                                f"SET {column._meta.name} = "
                                "{} WHERE id = {}"
                            ),
                            column.get_default_value(),
                            _id,
                        ).run()

        return response

    @property
    def querystrings(self) -> t.Sequence[QueryString]:
        if self._drop_table:
            return [QueryString(f'DROP TABLE "{self.table._meta.tablename}"')]

        query = f"ALTER TABLE {self.table._meta.tablename}"

        alterations = [
            i.querystring
            for i in itertools.chain(
                self._add, self._rename, self._drop, self._unique, self._null
            )
        ]

        if self.engine_type == "sqlite":
            # Can only perform one alter statement at a time.
            query += " {}"
            return [QueryString(query, i) for i in alterations]

        # Postgres can perform them all at once:
        query += ",".join([" {}" for i in alterations])

        return [QueryString(query, *alterations)]



import asyncio
import os
import sys

import click

from piccolo.migrations.tables import Migration
from .base import BaseMigrationManager


class BackwardsMigrationManager(BaseMigrationManager):
    def __init__(self, app_name: str, migration_name: str):
        self.migration_name = migration_name
        self.app_name = app_name

    def run(self):
        config_modules = self.get_config_modules()

        migration_modules = []

        for config_module in config_modules:
            if self.get_app_name(config_module) == self.app_name:
                migration_modules = self.get_migration_modules(
                    os.path.dirname(config_module.__file__)
                )
                break

        migration_ids = self.get_migration_ids(migration_modules)

        if self.migration_name not in migration_ids:
            print(
                f"Unrecognized migration name - must be one of {migration_ids}"
            )

        _continue = input(
            "About to undo the migrations - press y to continue."
        )
        if _continue == "y":
            print("Undoing migrations")

            _sorted_migration_ids = migration_ids[
                migration_ids.index(self.migration_name) :
            ]
            _sorted_migration_ids.reverse()

            already_ran = Migration.get_migrations_which_ran()

            for migration_name in _sorted_migration_ids:
                if migration_name not in already_ran:
                    print(
                        f"Migration {migration_name} hasn't run yet, can't "
                        "undo!"
                    )
                    sys.exit(1)

                print(f"Reversing {migration_name}")
                asyncio.run(
                    migration_modules[migration_name].backwards()
                )  # type: ignore

                Migration.delete().where(
                    Migration.name == migration_name
                ).run_sync()
        else:
            print("Not proceeding.")


@click.command()
@click.argument("app_name")
@click.argument("migration_name")
def backwards(migration_name: str, app_name: str):
    """
    Undo migrations up to a specific migrations.

    - make sure the migration name is valid
    - work out which to undo, and in which order
    - ask for confirmation
    - apply the undo operations one by one
    """
    manager = BackwardsMigrationManager(
        app_name=app_name, migration_name=migration_name
    )
    manager.run()



from piccolo.columns import Varchar, Boolean
from piccolo.table import Table


ID = "2019-11-14T21:52:21"


class BaseUser(Table, tablename="piccolo_user"):
    username = Varchar(length=100, unique=True)
    password = Varchar(length=255)
    email = Varchar(length=255, unique=True)
    active = Boolean(default=False)
    admin = Boolean(default=False)


async def forwards():
    await BaseUser.create_table().run()


async def backwards():
    await BaseUser.alter().drop_table().run()



import os

import click

from .base import BaseMigrationManager
from piccolo.migrations.tables import Migration


class CheckMigrationManager(BaseMigrationManager):
    def get_fixed_length_string(self, string: str, length=20) -> str:
        """
        Add spacing to the end of the string so it's a fixed length.
        """
        spacing = "".join([" " for i in range(20 - len(string))])
        fixed_length_string = f"{string}{spacing}"
        return fixed_length_string

    def run(self):
        print("Listing migrations ...")

        config_modules = self.get_config_modules()

        for config_module in config_modules:
            app_name = self.get_app_name(config_module)
            fixed_length_app_name = self.get_fixed_length_string(app_name)

            path = os.path.dirname(config_module.__file__)
            migration_modules = self.get_migration_modules(path)
            ids = self.get_migration_ids(migration_modules)
            for _id in ids:
                has_ran = (
                    Migration.exists()
                    .where(
                        (Migration.name == _id)
                        & (Migration.app_name == app_name)
                    )
                    .run_sync()
                )
                print(f"{fixed_length_app_name} | {_id} | {has_ran}")


@click.command()
def check():
    """
    Lists all migrations which have and haven't ran.
    """
    CheckMigrationManager().run()



from __future__ import annotations
import importlib
import os
import sys
from types import ModuleType
import typing as t

from piccolo.migrations.tables import Migration


class MigrationModule(ModuleType):
    @staticmethod
    async def forwards() -> None:
        pass

    @staticmethod
    async def backwards() -> None:
        pass


class ConfigModule(ModuleType):
    NAME: str
    DEPENDENCIES: t.List[str]


class BaseMigrationManager:
    def create_migration_table(self) -> bool:
        """
        Creates the migration table in the database. Returns True/False
        depending on whether it was created or not.
        """
        if not Migration.table_exists().run_sync():
            Migration.create_table().run_sync()
            return True
        return False

    def _deduplicate(
        self, config_modules: t.List[ConfigModule]
    ) -> t.List[ConfigModule]:
        """
        Remove all duplicates - just leaving the first instance.
        """
        return list(dict([(c, None) for c in config_modules]).keys())

    def import_config_modules(
        self, migrations: t.List[str]
    ) -> t.List[ConfigModule]:
        """
        Import all config modules, and all dependencies.
        """
        config_modules = []

        for config_module_path in migrations:
            try:
                config_module = t.cast(
                    ConfigModule, importlib.import_module(config_module_path)
                )
            except ImportError:
                raise Exception(f"Unable to import {config_module_path}")
            DEPENDENCIES = getattr(config_module, "DEPENDENCIES", [])
            dependency_config_modules = self.import_config_modules(
                DEPENDENCIES
            )
            config_modules.extend(dependency_config_modules + [config_module])

        return config_modules

    def get_config_modules(self) -> t.List[ConfigModule]:
        try:
            from piccolo_conf import MIGRATIONS
        except ImportError:
            raise Exception(
                "Unable to import MIGRATIONS from piccolo_conf - make sure "
                "it's in your path."
            )

        config_modules = self.import_config_modules(MIGRATIONS)

        # Now deduplicate any dependencies
        config_modules = self._deduplicate(config_modules)

        return config_modules

    def get_migration_modules(
        self, folder_path: str
    ) -> t.Dict[str, MigrationModule]:
        """
        Import the migration modules and store them in a dictionary.
        """
        migration_modules = {}

        sys.path.insert(0, folder_path)

        folder_contents = os.listdir(folder_path)
        excluded = ("__init__.py", "__pycache__", "config.py")
        migration_names = [
            i.split(".py")[0]
            for i in folder_contents
            if ((i not in excluded) and (not i.startswith(".")))
        ]

        modules: t.List[MigrationModule] = [
            t.cast(MigrationModule, importlib.import_module(name))
            for name in migration_names
        ]
        for m in modules:
            _id = getattr(m, "ID", None)
            if _id:
                migration_modules[_id] = m

        return migration_modules

    def get_migration_ids(
        self, migration_module_dict: t.Dict[str, MigrationModule]
    ) -> t.List[str]:
        return sorted(list(migration_module_dict.keys()))

    def get_app_name(self, config_module: ConfigModule) -> str:
        app_name = getattr(config_module, "NAME", "").strip()
        return app_name



from ..base import DBTestCase
from ..example_project.tables import Band


class TestUpdate(DBTestCase):
    def test_update(self):
        self.insert_rows()

        Band.update().values({Band.name: "Pythonistas3"}).where(
            Band.name == "Pythonistas"
        ).run_sync()

        response = (
            Band.select(Band.name)
            .where(Band.name == "Pythonistas3")
            .run_sync()
        )
        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas3"}])



import jinja2

TEMPLATE = """
ID = '{migration_id}'


async def forwards():
    pass


async def backwards():
    pass
"""


def render_template(**kwargs):
    template = jinja2.Template()

    return template.render(**kwargs)



import importlib
import os
import sys

import click

from piccolo.commands.playground import playground
from piccolo.commands.migration.new import new
from piccolo.commands.migration.backwards import backwards
from piccolo.commands.migration.forwards import forwards
from piccolo.commands.migration.check import check


@click.group()
def cli():
    pass


@cli.group("migration")
def migration():
    pass


migration.add_command(check)
migration.add_command(new)
migration.add_command(forwards)
migration.add_command(backwards)

cli.add_command(playground)


def main():
    # In case it's run from an entrypoint:
    sys.path.insert(0, os.getcwd())

    try:
        import piccolo_conf
    except ImportError:
        print("Can't import piccolo_conf - some commands may be missing.")
    else:
        COMMANDS = getattr(piccolo_conf, "COMMANDS", [])

        for command in COMMANDS:
            command_name = command.split(".")[-1]
            command_module = importlib.import_module(command)
            command_func = getattr(command_module, "command")
            cli.add_command(click.command(name=command_name)(command_func))

    cli()


if __name__ == "__main__":
    main()



from __future__ import annotations
import datetime
import os
import typing as t
from types import ModuleType

import click

from piccolo.migrations.template import TEMPLATE


MIGRATION_MODULES: t.Dict[str, ModuleType] = {}


def _create_migrations_folder(migrations_path: str) -> bool:
    """
    Creates the folder that migrations live in. Returns True/False depending
    on whether it was created or not.
    """
    if os.path.exists(migrations_path):
        return False
    else:
        os.mkdir(migrations_path)
        for filename in ("__init__.py", "config.py"):
            with open(os.path.join(migrations_path, filename), "w"):
                pass
        return True


def _create_new_migration(migrations_path) -> None:
    """
    Creates a new migration file on disk.
    """
    _id = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    path = os.path.join(migrations_path, f"{_id}.py")
    with open(path, "w") as f:
        f.write(TEMPLATE.format(migration_id=_id))


###############################################################################


@click.option(
    "-p",
    "--path",
    multiple=False,
    help="The parent of the migrations folder e.g. ./my_app",
)
@click.command()
def new(create: t.Tuple[str], path: str):
    """
    Creates a new file like migrations/2018-09-04T19:44:09.py
    """
    print("Creating new migration ...")

    root_dir = path if path else os.getcwd()
    migrations_path = os.path.join(root_dir, "migrations")

    _create_migrations_folder(migrations_path)
    _create_new_migration(migrations_path)
